import unittest

import redino
from tests.redino.test_shared import prepare_test


class MigratedItem(redino.Entity):
    data: str


redino.Entity.attributes(MigratedItem, {"data": str})


class TestMigration(unittest.TestCase):
    """
    Tests if creating model works
    """

    def test_migration(self) -> None:
        # we simulate a cold start, but with already some data
        # inside
        prepare_test()

        @redino.connect
        @redino.transactional
        def persist_an_item():
            item = MigratedItem().rd_persist()
            item.data = "test data"

        persist_an_item()

        # now here goes the cold start
        redino.model.dbversion = redino.model.VERSION_UKNOWN

        # we need to call the redino.model.setup before any
        # redino operations
        @redino.connect
        def read_migrated_item():
            items = redino.Entity.fetch_all(MigratedItem)
            self.assertTrue(items)
            self.assertEqual(1, len(items))
            self.assertEqual("test data", items[0].data)

        read_failed = True
        try:
            read_migrated_item()
            read_failed = False
        except Exception:
            pass

        self.assertTrue(read_failed)

        # we should already have some version inside it, so we
        # ensure that the migration is the previous version
        def migration_check(version):
            self.assertEqual("1", version)

        redino.model.setup(version="2.custom", migrate=migration_check)

        read_migrated_item()

        @redino.connect
        def check_version_data():
            model_definitions = redino.Entity.fetch_all(redino.model.RedinoModel)
            self.assertTrue(model_definitions)
            self.assertEqual(1, len(model_definitions))
            self.assertEqual("2.custom", model_definitions[0].version)

        check_version_data()



import redis
import threading

_redis_thread_instance = threading.local()


def redis_instance() -> redis.client.Redis:
    if not _redis_thread_instance.instance:
        raise Exception("This works only inside a @redino.connect function")

    return _redis_thread_instance.instance



import unittest
from typing import Set

import redino
from redino import RedinoSet
from tests.redino.test_shared import prepare_test


class TestModel(unittest.TestCase):
    """
    Tests if creating model works
    """

    @classmethod
    def setUpClass(cls) -> None:
        prepare_test()

    @redino.connect
    def test_set(self):
        s = TestModel.create_test_set()

        try:
            self.assertEqual(3, len(s))
        finally:
            s.rd_delete()

    @redino.connect
    def test_contains(self):
        s = TestModel.create_test_set()

        try:
            self.assertTrue("b" in s)
            self.assertTrue("y" not in s)
        finally:
            s.rd_delete()

    @redino.connect
    def test_delete_and_iterator(self):
        s = TestModel.create_test_set()

        try:
            self.assertEqual(3, len(s))
            s.remove("b")

            self.assertEqual(2, len(s))
            self.assertEqual(
                {"a", "c"},
                set(s),
            )
        finally:
            s.rd_delete()

    @staticmethod
    def create_test_set() -> RedinoSet:
        s = RedinoSet(_type=Set[str], _id="testset").rd_persist()

        s.add("a")
        s.add("b")
        s += "c"

        return s



import inspect
from typing import Any, Union

import redino.redino_entity
import redino.redino_entity
import redino.redino_item

RedisBytes = Union[bytes, int, str, None]
RedinoNative = Union[redino.redino_item.RedinoItem, bytes, int, str, None]


class DataConverter:
    def __init__(self, _type: Any) -> None:
        self._type = _type

    def from_data(self, data: Any) -> RedinoNative:
        if self._type is str and isinstance(data, str):
            return data

        if self._type is int and isinstance(data, int):
            return data

        if isinstance(data, redino.redino_entity.RedinoItem):
            return data

        if self._type.__origin__ is list and isinstance(data, list):
            list_result = redino.redino_list.RedinoList(_type=self._type)
            list_result.extend(data)

            return list_result

        if self._type.__origin__ is set and isinstance(data, set):
            set_result = redino.redino_set.RedinoSet(_type=self._type)
            set_result.update(data)

            return set_result

        if self._type.__origin__ is dict and isinstance(data, dict):
            dict_result = redino.redino_dict.RedinoDict(_type=self._type)
            dict_result.update(data)

            return dict_result

        raise Exception(f"Unsupported type: {self._type} for data {type(data)}")

    def from_bytes(self, data: bytes) -> RedinoNative:
        if data is None:
            return None

        if self._type is str:
            return data.decode("utf-8")

        if self._type is int:
            return int(data)

        if inspect.isclass(self._type) and issubclass(
            self._type, redino.redino_entity.Entity
        ):
            return self._type(_id=data.decode("utf-8"))

        if self._type.__origin__ is list:
            return redino.redino_list.RedinoList(
                _type=self._type, _id=data.decode("utf-8")
            )

        if self._type.__origin__ is set:
            return redino.redino_set.RedinoSet(
                _type=self._type, _id=data.decode("utf-8")
            )

        if self._type.__origin__ is dict:
            return redino.redino_dict.RedinoDict(
                _type=self._type, _id=data.decode("utf-8")
            )

        raise Exception("Unsupported type: " % self._type)

    def data_to_bytes(self, data: Any) -> RedisBytes:
        item = self.from_data(data)

        if isinstance(item, redino.redino_item.RedinoItem):
            return item._rd_self_id

        return item

    def native_to_bytes(self, native: RedinoNative) -> RedisBytes:
        if isinstance(native, redino.redino_item.RedinoItem):
            return native._rd_self_id

        return native



from typing import Any, Optional, TypeVar

from redino import redis_instance
from redino._redis_scan import RedisIterable, RedisScan
from redino.data_converter import DataConverter, RedinoNative
from redino.redino_item import RedinoItem

_T = RedinoNative
_S = TypeVar("_S", bound="RedinoSet")


class RedinoSet(RedinoItem):
    def __init__(self, _type: Any, _id: Optional[str] = None) -> None:
        super(RedinoSet, self).__init__(_id=_id)

        # FIXME: use a converter cache for the types
        self._rd_converter = DataConverter(_type=_type.__args__[0])

    def rd_persist(self: _S) -> _S:
        return self

    def rd_delete(self: _S) -> None:
        self.clear()

    def add(self: _S, item: _T):
        redis_instance().execute_command(
            "sadd", self._rd_self_id, self._rd_converter.data_to_bytes(item)
        )

    def clear(self):
        redis_instance().execute_command("del", self._rd_self_id)

    def discard(self, item: _T):
        redis_instance().execute_command(
            "srem",
            self._rd_self_id,
            self._rd_converter.data_to_bytes(item),
        )

    def remove(self, item: _T):
        result = int(
            redis_instance().execute_command(
                "srem",
                self._rd_self_id,
                self._rd_converter.data_to_bytes(item),
            )
        )

        if not result:
            raise ValueError(f"No such item {str(item)} in set.")

    def update(self, other_set):
        for item in other_set:
            self.add(item)

    def __contains__(self, item) -> bool:
        return redis_instance().execute_command(
            "sismember",
            self._rd_self_id,
            self._rd_converter.data_to_bytes(item),
        )

    def __iter__(self, *args, **kwargs):  # real signature unknown
        """ Implement iter(self). """
        return RedisIterable(self, RedisSetIterator).__iter__()

    def __len__(self):  # real signature unknown
        return int(
            redis_instance().execute_command(
                "scard",
                self._rd_self_id,
            )
        )

    def __iadd__(self: _S, other: _T) -> _S:
        self.add(other)
        return self


class RedisSetIterator:
    def __init__(self, s: RedinoSet) -> None:
        self._scan = RedisScan(s, "sscan")
        self._s = s

    def __next__(self) -> _T:
        return self._s._rd_converter.from_bytes(self._scan.__next__())



import unittest
from typing import List

import redino
from tests.redino.test_shared import prepare_test


class TestList(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        prepare_test()

    @redino.connect
    def test_list_len(self):
        l = self.create_test_list()

        try:
            self.assertEqual(3, len(l))
        finally:
            l.rd_delete()

    @redino.connect
    def test_list_index(self):
        l = self.create_test_list()
        try:
            self.assertEqual("x", l[0])
            self.assertEqual("y", l[1])
            self.assertEqual("z", l[2])
        finally:
            l.rd_delete()

    @redino.connect
    def test_remove_by_index(self):
        l = self.create_test_list()
        try:
            del l[1]

            self.assertEqual(2, len(l))
            self.assertEqual("x", l[0])
            self.assertEqual("z", l[1])
        finally:
            l.rd_delete()

    @redino.connect
    def test_contains(self):
        l = self.create_test_list()
        try:
            self.assertTrue("y" in l)
            self.assertTrue("b" not in l)
        finally:
            l.rd_delete()

    @redino.connect
    def test_clear(self):
        l = self.create_test_list()
        try:
            self.assertEqual(3, len(l))
            l.clear()
            self.assertEqual(0, len(l))
        finally:
            l.rd_delete()

    @redino.connect
    def test_iteration(self):
        l = self.create_test_list()
        try:
            for f in l:
                print(f)
        finally:
            l.rd_delete()

    @redino.connect
    def test_left_right_push(self):
        l = self.create_test_list()
        try:
            l.clear()
            l.right_push("b")
            l.right_push("c")
            l.left_push("a")

            self.assertEqual(3, len(l))
            self.assertEqual("a", l[0])
            self.assertEqual("b", l[1])
            self.assertEqual("c", l[2])

            self.assertEqual("a", l.left_pop())
            self.assertEqual("c", l.right_pop())

            self.assertEqual(1, len(l))
            self.assertEqual("b", l[0])
        finally:
            l.rd_delete()

    def create_test_list(self):
        l = redino.RedinoList(
            _id="_testlist",
            _type=List[str],
        ).rd_persist()

        l.clear()

        l.append("x")
        l.append("y")
        l += "z"

        return l



import abc
import inspect
import uuid
from typing import Optional, Any, TypeVar

_S = TypeVar("_S")


def class_name(instance: Any) -> str:
    t = instance if inspect.isclass(instance) else type(instance)

    return f"{t.__name__}"


class RedinoItem(metaclass=abc.ABCMeta):
    """
    Marker class to have easier isinstance checks for the
    DataConverter.
    """

    def __init__(self, _id: Optional[str] = None) -> None:
        if _id is not None:
            self._rd_self_id = _id
        else:
            self._rd_self_id = f"{class_name(self)}:{str(uuid.uuid4())}"

    @abc.abstractmethod
    def rd_persist(self: _S) -> _S:
        pass

    @abc.abstractmethod
    def rd_delete(self: _S) -> None:
        pass



import functools
import unittest
from typing import List, TypeVar, Callable, Dict, Set

import redino
from redino import RedinoSet
from redino.data_converter import DataConverter
from tests.redino.test_shared import prepare_test

T = TypeVar("T")


def expect_fail(f: Callable[..., None]) -> Callable[..., None]:
    @functools.wraps(f)
    def wrapper(*args, **kw) -> None:
        is_failed = False

        try:
            f(*args, **kw)
        except Exception as e:
            is_failed = True

        if not is_failed:
            raise Exception("Code didn't fail as expected")

    return wrapper


class TestDataConverter(unittest.TestCase):
    """
    Tests if conversion handles correctly
    """

    @classmethod
    def setUpClass(cls) -> None:
        prepare_test()

    @redino.connect
    def test_nested_lists(self):
        t = List[List[str]]
        converter = DataConverter(t)

        # create a nested redino list of items
        l = converter.from_data([[]])

        self.assertEqual(1, len(l))
        self.assertEqual(0, len(l[0]))

        @expect_fail
        def assign_wrong_type_str_to_list_fails():
            l.append("abc")

        assign_wrong_type_str_to_list_fails()

        @expect_fail
        def assign_wrong_type_list_to_str_fails():
            l[0].append([])

        assign_wrong_type_list_to_str_fails()

    @redino.connect
    def test_dictionaries(self):
        t = Dict[str, Dict[str, str]]
        converter = DataConverter(t)

        # create a nested redino list of items
        d = converter.from_data(
            {
                "x": {
                    "y": "abc",
                    "z": "123",
                }
            }
        )

        self.assertTrue(1, len(d))
        self.assertTrue(2, len(d["x"]))

        @expect_fail
        def assign_wrong_type_on_keys_should_fail():
            d[3] = {"a": "b"}

        assign_wrong_type_on_keys_should_fail()

        @expect_fail
        def assign_wrong_type_on_values_should_fail():
            d["a"] = {"a", "b"}  # this is a set, not a dict

        assign_wrong_type_on_values_should_fail()

    @redino.connect
    def test_set_assignment(self):
        t = Set[str]
        converter = DataConverter(t)

        # create a nested redino list of items
        s = converter.from_data({"a", "b", "c"})

        self.assertTrue(isinstance(s, RedinoSet))
        self.assertEqual(3, len(s))

        s.add("d")
        self.assertEqual(4, len(s))

        @expect_fail
        def assign_wrong_type_on_set_should_fail():
            s.add(3)

        assign_wrong_type_on_set_should_fail()


if __name__ == "__main__":
    unittest.main()



from typing import Any

from redino import redis_instance
from redino.redino_item import RedinoItem


class RedisIterable:
    def __init__(self, d: RedinoItem, iterator_class: Any) -> None:
        self._d = d
        self._iterator_class = iterator_class

    def __iter__(self):
        return self._iterator_class(self._d)


class RedisScan:
    def __init__(self, d: RedinoItem, command: str) -> None:
        self._d = d
        self._current_index = -1
        self._command = command

        self._read_redis_cursor()

    def __next__(self) -> Any:
        try:
            return self._cursor.__next__()
        except StopIteration:
            self._read_redis_cursor()
            return self._cursor.__next__()

    def _read_redis_cursor(self):
        # if the next index is 0, the redis scan is complete
        if self._current_index == 0:
            raise StopIteration()

        # this is why we mark the first index as -1 to start the scan
        # so we don't confuse it with completion
        if self._current_index == -1:
            self._current_index = 0

        redis_cursor = redis_instance().execute_command(
            self._command,
            self._d._rd_self_id,
            self._current_index,
        )

        self._current_index = redis_cursor[0]

        if hasattr(redis_cursor[1], "items"):
            self._cursor = redis_cursor[1].items().__iter__()
        else:
            self._cursor = redis_cursor[1].__iter__()



from typing import TypeVar, Iterable, Any, Optional, Iterator

from redino._redis_instance import redis_instance
from redino.data_converter import DataConverter, RedinoNative
from redino.redino_item import RedinoItem

_T = RedinoNative


class RedinoList(RedinoItem):
    """
    This behaves a lot like a linked list deque, so using indexed
    access it's a terrible idea. Has blocking support for fetching
    items.
    """

    def __init__(self, _type: Any, _id: Optional[str] = None) -> None:
        super(RedinoList, self).__init__(_id=_id)

        # FIXME: use a converter cache for the types
        self._rd_converter = DataConverter(_type=_type.__args__[0])

    def rd_persist(self: "RedinoList") -> "RedinoList":
        # there's nothing to persist when created
        return self

    def rd_delete(self) -> None:
        self.clear()

    def __contains__(self, item: _T) -> bool:
        return self.index(item) >= 0

    def __len__(self) -> int:
        result = redis_instance().execute_command("llen", self._rd_self_id)
        return int(result)

    def __setitem__(self, index: int, data: _T) -> None:
        redis_instance().execute_command(
            "lset", self._rd_self_id, index, self._rd_converter.data_to_bytes(data)
        )

    def __getitem__(self: "RedinoList", index: int) -> _T:
        data = redis_instance().execute_command("lindex", self._rd_self_id, index)

        return self._rd_converter.from_bytes(data)

    def __delitem__(self, i: int) -> None:
        self[i] = "__remove_me"
        self.remove("__remove_me")

    def __iadd__(self: "RedinoList", other: _T) -> "RedinoList":
        self.append(other)
        return self

    def __iter__(self) -> Iterable[_T]:
        return ListIterator(self)

    def append(self, other: _T):
        redis_instance().execute_command(
            "rpush", self._rd_self_id, self._rd_converter.data_to_bytes(other)
        )

    def pop(self) -> _T:
        data = redis_instance().execute_command("lpop", self._rd_self_id)
        return self._rd_converter.from_bytes(data)

    def left_push(self: "RedinoList", item: _T) -> None:
        data = self._rd_converter.data_to_bytes(item)
        redis_instance().execute_command("lpush", self._rd_self_id, data)

    def right_push(self: "RedinoList", item: _T) -> None:
        data = self._rd_converter.data_to_bytes(item)
        redis_instance().execute_command("rpush", self._rd_self_id, data)

    def left_pop(self: "RedinoList") -> _T:
        data = redis_instance().execute_command("lpop", self._rd_self_id)
        return self._rd_converter.from_bytes(data)

    def right_pop(self: "RedinoList") -> _T:
        data = redis_instance().execute_command("rpop", self._rd_self_id)
        return self._rd_converter.from_bytes(data)

    def blocking_left_push(self: "RedinoList", item: _T) -> None:
        data = self._rd_converter.data_to_bytes(item)
        redis_instance().execute_command("blpush", self._rd_self_id, data)

    def blocking_right_push(self: "RedinoList", item: _T) -> None:
        data = self._rd_converter.data_to_bytes(item)
        redis_instance().execute_command("brpush", self._rd_self_id, data)

    def blocking_left_pop(self: "RedinoList") -> _T:
        data = redis_instance().execute_command("blpop", self._rd_self_id)
        return self._rd_converter.from_bytes(data)

    def blocking_right_pop(self: "RedinoList") -> _T:
        data = redis_instance().execute_command("brpop", self._rd_self_id)
        return self._rd_converter.from_bytes(data)

    def remove(self, item: _T) -> None:
        item_bytes = self._rd_converter.data_to_bytes(item)
        redis_instance().execute_command(
            "lrem", self._rd_self_id, 1, item_bytes  # key  # count  # element
        )

    def clear(self) -> None:
        redis_instance().execute_command("del", self._rd_self_id)

    def index(self, item: _T, *args: Any) -> int:
        result = redis_instance().execute_command(
            "lpos",
            self._rd_self_id,
            self._rd_converter.data_to_bytes(item),
        )

        if result is None:
            return -1

        return result

    def extend(self, other: Iterable[_T]) -> None:
        for item in other:
            self.append(item)


class ListIterator(Iterator[_T]):
    def __init__(self, l: RedinoList) -> None:
        self._list = l
        self._len = len(l)
        self._index = 0

    def __next__(self) -> _T:
        if self._index < self._len:
            result = self._list[self._index]
            self._index += 1

            return result

        raise StopIteration



import unittest

import redino
from redino import redis_instance
from tests.redino.test_shared import prepare_test


class TestRedisReadWrite(unittest.TestCase):
    """
    Test the model if it can create entries in redis
    """

    @classmethod
    def setUpClass(cls) -> None:
        prepare_test()

    def test_regular_writing(self):
        @redino.connect
        @redino.transactional
        def set_values():
            r = redino.redis_instance()

            r.hset("a", "key", "3")
            r.hset("b", "key", "3")

        @redino.connect
        def read_values():
            r = redino.redis_instance()

            self.assertEqual("3", r.hget("a", "key").decode("utf-8"))
            self.assertEqual("3", r.hget("b", "key").decode("utf-8"))

        @redino.connect
        def clean_values():
            redis_instance().execute_command("del", "a")
            redis_instance().execute_command("del", "b")

        try:
            set_values()
            read_values()
        finally:
            clean_values()

    def test_exception_rolls_back(self):
        @redino.connect
        @redino.transactional
        def set_values():
            r = redino.redis_instance()

            r.hset("ROLL", "key", "3")
            raise Exception("ded")

        @redino.connect
        def read_values():
            r = redino.redis_instance()

            self.assertIsNone(r.hget("ROLL", "key"))

        try:
            set_values()
        except Exception:
            pass

        read_values()


if __name__ == "__main__":
    unittest.main()



import unittest
from typing import Dict, Set, List

import redino
from tests.redino.test_shared import prepare_test


class Item(redino.Entity):
    name: str
    count: int
    d: Dict[int, int]
    s: Set[str]
    l: List[int]

    attr = {
        "name": str,
        "count": int,
        "d": Dict[int, int],
        "s": Set[str],
        "l": List[int],
    }


class TestModel(unittest.TestCase):
    """
    Tests if creating model works
    """

    @classmethod
    def setUpClass(cls) -> None:
        prepare_test()

    def test_model(self):
        """
        Tests a model creation
        """

        @redino.connect
        def clear_values():
            for it in redino.Entity.fetch_all(Item):
                if it.d:
                    it.d.rd_delete()

                if it.s:
                    it.s.rd_delete()

                if it.l:
                    it.l.rd_delete()

                it.rd_delete()

        @redino.connect
        @redino.transactional
        def set_values():
            item = Item().rd_persist()
            item.name = "wut"
            item.count = 5

        @redino.connect
        def read_values():
            items: List[Item] = redino.Entity.fetch_all(Item)
            self.assertEqual(1, len(items))

            self.assertEqual("wut", items[0].name)
            self.assertTrue(isinstance(items[0].name, str))
            self.assertEqual(5, items[0].count)
            self.assertTrue(isinstance(items[0].count, int))

        try:
            clear_values()
            set_values()
            read_values()
        finally:
            clear_values()


if __name__ == "__main__":
    unittest.main()



import functools
from typing import Callable, TypeVar, Optional

import redino
import redino.redino_entity
from redino._redis_instance import _redis_thread_instance

T = TypeVar("T")


# The version means several things:
# "<unknown>"                 - if the setup() function was not yet called
# None                        - if the setup() function is actively running
# Any other non-empty string  - if the setup() was successful, with the active model version

VERSION_UKNOWN = "<unknown>"
dbversion = VERSION_UKNOWN


class RedinoModel(redino.redino_entity.Entity):
    """
    This is the persisted entity that holds the version
    """

    version: str


redino.redino_entity.Entity.attributes(
    RedinoModel,
    {
        "version": str,
    },
)


def _setup_connect(f: Callable[..., T]) -> Callable[..., T]:
    @functools.wraps(f)
    def wrapper(*args, **kw) -> T:
        with redino._redis_pool().client() as redis_client:  # type: ignore
            try:
                _redis_thread_instance.instance = redis_client
                return f(*args, **kw)
            finally:
                _redis_thread_instance.instance = None

    return wrapper


@_setup_connect
def setup(*, migrate: Callable[[Optional[str]], None], version: str = "") -> None:
    """
    A function to allow migration of data from one
    format to another.
    """
    global dbversion

    if not version:
        raise Exception("version should be a non-empty string.")

    current_model_version = None

    redino_model = redino.redino_entity.Entity.fetch_first(RedinoModel)
    if redino_model:
        current_model_version = redino_model.version

    if current_model_version == version:
        dbversion = version
        return

    migrate(current_model_version)

    # if we don't have a model we'll create the model as well
    if not redino_model:
        redino_model = RedinoModel().rd_persist()

    redino_model.version = version
    dbversion = version



from typing import Optional
import unittest

import redino.model

import redino.model
import redino._redis_instance

test = unittest.TestCase()

#
# Since we're enforcing the redino.model.setup() call, we need
# to ensure that in-between tests everything is nuked, and
# the model has a correct version.
#


def empty_redino_model_setup(version: Optional[str]):
    test.assertIsNone(version)


@redino.model._setup_connect
def destroy_all_data():
    redino._redis_instance.redis_instance().flushall()
    redino.model.dbversion = redino.model.VERSION_UKNOWN  # the setup was not yet called


def prepare_test():
    destroy_all_data()
    redino.model.setup(version="1", migrate=empty_redino_model_setup)



from typing import TypeVar, Any, Optional, Dict, Tuple, Iterable

import redino.data_converter
import redino.redino_item
from redino import redis_instance
from redino._redis_scan import RedisScan, RedisIterable

_S = TypeVar("_S", bound="RedinoDict")
_K = redino.data_converter.RedinoNative
_V = redino.data_converter.RedinoNative


class RedinoDict(redino.redino_item.RedinoItem):
    def __init__(self: _S, _type: Any, _id: Optional[str] = None) -> None:
        super(RedinoDict, self).__init__(_id=_id)

        # FIXME: use a converter cache for the types
        self._rd_converter_key = redino.data_converter.DataConverter(
            _type=_type.__args__[0]
        )
        self._rd_converter_value = redino.data_converter.DataConverter(
            _type=_type.__args__[1]
        )

    def rd_persist(self: _S) -> _S:
        return self

    def rd_delete(self) -> None:
        self.clear()

    def clear(self):
        redis_instance().execute_command("del", self._rd_self_id)

    def get(self, key: _K) -> _V:
        data = redis_instance().execute_command(
            "hget",
            self._rd_self_id,
            self._rd_converter_key.data_to_bytes(key),
        )

        return self._rd_converter_value.from_bytes(data)

    def items(self) -> Iterable[Tuple[_K, _V]]:
        return RedisIterable(self, RedisDictItems)

    def keys(self):
        return RedisIterable(self, RedisDictKeys)

    def pop(self, k, d=None):
        """
        D.pop(k[,d]) -> v, remove specified key and return the corresponding value.
        If key is not found, d is returned if given, otherwise KeyError is raised
        """
        pass

    def popitem(self):  # real signature unknown; restored from __doc__
        """
        D.popitem() -> (k, v), remove and return some (key, value) pair as a
        2-tuple; but raise KeyError if D is empty.
        """
        pass

    def setdefault(self, *args, **kwargs):  # real signature unknown
        pass

    def update(self, m: Dict[_K, _V]):
        for k, v in m.items():
            self[k] = v

    def values(self):
        return RedisIterable(self, RedisDictValues)

    def __contains__(self, key: _K) -> bool:
        return redis_instance().execute_command(
            "hexists", self._rd_self_id, self._rd_converter_key.data_to_bytes(key)
        )

    def __delitem__(self, key: _K):
        redis_instance().execute_command(
            "hdel",
            self._rd_self_id,
            self._rd_converter_key.data_to_bytes(key),
        )

    def __getitem__(self, key: _K) -> Optional[_V]:
        data = redis_instance().execute_command(
            "hget",
            self._rd_self_id,
            self._rd_converter_key.data_to_bytes(key),
        )

        return self._rd_converter_value.from_bytes(data)

    def __iter__(self):
        return self.keys()

    def __len__(self) -> int:
        return int(
            redis_instance().execute_command(
                "hlen",
                self._rd_self_id,
            )
        )

    def __setitem__(self, key: _K, value: _V):
        redis_instance().execute_command(
            "hset",
            self._rd_self_id,
            self._rd_converter_key.data_to_bytes(key),
            self._rd_converter_value.data_to_bytes(value),
        )


def convert_data(d: RedinoDict, data: Tuple[Any, Any]) -> Tuple[_K, _V]:
    return (
        d._rd_converter_key.from_bytes(data[0]),
        d._rd_converter_value.from_bytes(data[1]),
    )


class RedisDictItems:
    def __init__(self, d: RedinoDict) -> None:
        self._scan = RedisScan(d, "hscan")
        self._d = d

    def __next__(self) -> Tuple[_K, _V]:
        return convert_data(self._d, self._scan.__next__())


class RedisDictKeys:
    def __init__(self, d: RedinoDict) -> None:
        self._iterator = RedisDictItems(d)

    def __next__(self) -> _K:
        return self._iterator.__next__()[0]


class RedisDictValues:
    def __init__(self, d: RedinoDict) -> None:
        self._iterator = RedisDictItems(d)

    def __next__(self) -> _K:
        return self._iterator.__next__()[1]



import unittest
import uuid
from typing import Dict

import redino
from redino.redino_dict import RedinoDict
from tests.redino.test_shared import prepare_test


class TestDictionary(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        prepare_test()

    @redino.connect
    def test_dict_items(self):
        d = TestDictionary.create_dictionary()
        try:
            index = 0
            for k, v in d.items():
                if index == 0:
                    self.assertEqual("x", k)
                    self.assertEqual("a", v)
                elif index == 1:
                    self.assertEqual("y", k)
                    self.assertEqual("b", v)
                elif index == 2:
                    self.assertEqual("z", k)
                    self.assertEqual("c", v)
                elif index == 3:
                    raise Exception("There should be only 3 items")

                index += 1
        finally:
            d.rd_delete()

    def test_large_items(self):
        created_dict = dict()
        control_dict = dict()

        @redino.connect
        def create_large_dict():
            d = RedinoDict(_id="testdictlarge", _type=Dict[str, str]).rd_persist()

            for i in range(1000):
                key = str(uuid.uuid4())
                value = str(uuid.uuid4())

                created_dict[key] = value
                d[key] = value

        @redino.connect
        def read_dictionary():
            d = RedinoDict(_id="testdictlarge", _type=Dict[str, str])

            for k, v in d.items():
                control_dict[k] = v

        @redino.connect
        def delete_dict():
            RedinoDict(_id="testdictlarge", _type=Dict[str, str]).rd_delete()

        try:
            create_large_dict()
            read_dictionary()

            self.assertEqual(created_dict, control_dict)
        finally:
            delete_dict()

    @redino.connect
    def test_contains(self):
        d = TestDictionary.create_dictionary()

        try:
            self.assertTrue("x" in d)
            self.assertTrue("a" not in d)
        finally:
            d.rd_delete()

    @redino.connect
    def test_values_iterator(self):
        d = TestDictionary.create_dictionary()

        try:
            self.assertEqual({"a", "b", "c"}, set(d.values()))
        finally:
            d.rd_delete()

    @redino.connect
    def test_keys_iterator(self):
        d = TestDictionary.create_dictionary()

        try:
            self.assertEqual({"x", "y", "z"}, set(d.keys()))
        finally:
            d.rd_delete()

    @redino.connect
    def test_delete_item(self):
        d = TestDictionary.create_dictionary()

        try:
            del d["y"]

            self.assertEqual({"x", "z"}, set(d.keys()))
        finally:
            d.rd_delete()

    @redino.connect
    def test_len(self):
        d = TestDictionary.create_dictionary()

        try:
            self.assertEqual(3, len(d))
        finally:
            d.rd_delete()

    @staticmethod
    def create_dictionary() -> RedinoDict:
        d = RedinoDict(_id="testdict", _type=Dict[str, str]).rd_persist()

        d["x"] = "a"
        d["y"] = "b"
        d["z"] = "c"

        return d



import functools
import logging
from enum import Enum
from typing import Callable, TypeVar, Optional
import os

import redis.client

import redino.model
from redino._redis_instance import _redis_thread_instance, redis_instance

LOG = logging.getLogger(__name__)

T = TypeVar("T")
_redis_pool_instance: Optional[redis.client.Redis] = None


class Operation(Enum):
    CREATE = "CREATE"
    UPDATE = "UPDATE"
    DELETE = "DELETE"


def _redis_pool() -> redis.client.Redis:
    global _redis_pool_instance

    if not _redis_pool_instance:
        _redis_pool_instance = redis.client.Redis(
            host=os.environ.get("REDINO_HOST", "localhost"),
            port=int(os.environ.get("REDINO_PORT", "6379")),
            db=int(os.environ.get("REDINO_DB", "0")),
            password=os.environ.get("REDINO_PASSWORD", None),
        )

    return _redis_pool_instance


def watch(entity_class) -> Callable[..., Callable[..., T]]:
    def wrapper_builder(f: Callable[..., T]) -> Callable[..., T]:
        @functools.wraps(f)
        def wrapper(*args, **kw) -> T:
            return f(*args, **kw)

        return wrapper

    return wrapper_builder


def transactional(f: Callable[..., T]) -> Callable[..., T]:
    @functools.wraps(f)
    def wrapper(*args, **kw) -> T:
        r = redis_instance()

        r.execute_command("MULTI")
        try:
            result = f(*args, **kw)
            r.execute_command("EXEC")

            return result
        except Exception as e:
            r.execute_command("DISCARD")
            LOG.error("DISCARD redis changes, due to {err}", err=str(e))
            raise e

    return wrapper


def connect(f: Callable[..., T]) -> Callable[..., T]:
    @functools.wraps(f)
    def wrapper(*args, **kw) -> T:
        if (
            not redino.model.dbversion
            or redino.model.dbversion == redino.model.VERSION_UKNOWN
        ):
            raise Exception(
                "redino.model.setup(version, migrate) was not called, "
                "or model version is invalid. The version is the number that "
                "this application model defines, and the migrate() will receive "
                "the version stored in the datastore. If the code of migrate is "
                "successful, the datastore version is updated to the model "
                "version."
            )

        with _redis_pool().client() as redis_client:  # type: ignore
            try:
                _redis_thread_instance.instance = redis_client
                return f(*args, **kw)
            finally:
                _redis_thread_instance.instance = None

    return wrapper


from redino.redino_entity import Entity
from redino.redino_list import RedinoList
from redino.redino_dict import RedinoDict
from redino.redino_set import RedinoSet



from typing import List, Any, Optional, TypeVar, Dict, Type

from redino._redis_instance import redis_instance
from redino.redino_item import RedinoItem, class_name
from redino.data_converter import DataConverter, RedinoNative

T = TypeVar("T", bound="Entity")


class Entity(RedinoItem):
    """
    Entities are objects that can be persisted in Redis.

    They are backed by a HSET (HashSet) Redis object, and are
    identified by: `TYPE:id` key names. If they contain a list,
    set or dict, the backing property ID is: `TYPE:id:field_name`,
    and they return one of the `RedinoList`, `RedinoSet` or
    `RedinoDict` respectively.
    """

    def __init__(self, _id: Optional[str] = None) -> None:
        super(Entity, self).__init__(_id=_id)
        self._rd_cache: Dict[str, RedinoNative] = dict()

    def rd_persist(self: T) -> T:
        redis_instance().hset(class_name(self), self._rd_self_id, "1")
        return self

    def rd_delete(self) -> None:
        # FIXME: iterate over the `attr` and delete them, if the
        # items are being owned
        redis_instance().execute_command("hdel", class_name(self), self._rd_self_id)
        redis_instance().execute_command("del", self._rd_self_id)

    def __getattr__(self, key: str) -> Any:
        if key.startswith("_rd_"):
            return super(Entity, self).__getattribute__(key)

        if key not in type(self).attr:  # type: ignore
            raise Exception(
                f"No attribute {key} in {self._rd_self_id}. "  # type: ignore
                f"Only {type(self).attr} are known."
            )

        if key in self._rd_cache:
            return self._rd_cache[key]

        definition = type(self).attr[key]  # type: ignore
        # FIXME: cache converters
        converter = DataConverter(_type=definition)
        data = converter.from_bytes(redis_instance().hget(self._rd_self_id, key))

        self._rd_cache[key] = data

        return data

    def __setattr__(self, key, value):
        if key.startswith("_rd_"):
            return super(Entity, self).__setattr__(key, value)

        if key not in type(self).attr:
            raise Exception(
                f"No attribute {key} in {self._rd_self_id}. "
                f"Only {type(self).attr} are known."
            )

        definition = type(self).attr[key]

        # FIXME: cache converters
        converter = DataConverter(_type=definition)

        # NOTE: deletion is not handled here since we cannot
        # guarantee usage from other objects.
        native = converter.from_data(value)

        redis_instance().hset(self._rd_self_id, key, converter.native_to_bytes(native))

        self._rd_cache[key] = native

    @staticmethod
    def fetch_all(t: Type[T]) -> List[T]:
        """
        Fetches all the instances of the given type.
        """
        # FIXME: return an iterator
        # this gives us the IDs
        items = redis_instance().hgetall(class_name(t))
        return [t(_id=id.decode("utf-8")) for id in items]

    @staticmethod
    def fetch_first(t: Type[T]) -> Optional[T]:
        """
        Fetches a single item of a given type.
        """
        for item in Entity.fetch_all(t):
            return item

        return None

    @staticmethod
    def attributes(t: Type[T], attr: Dict[str, Any]) -> None:
        t.attr = attr  # type: ignore



import unittest
from typing import Optional, List

import redino
from redino import Entity
from tests.redino.test_shared import prepare_test


class Folder(redino.Entity):
    parent: Optional["Folder"]
    name: str
    folders: List["Folder"]
    files: List["File"]

    def rd_delete(self) -> None:
        if self.folders:
            for folder in self.folders:
                folder.rd_delete()

            self.folders.clear()

        # we call the remove of the parent after all the other removes
        super(Folder, self).rd_delete()


class File(redino.Entity):
    parent: Folder
    name: str
    content: str


Entity.attributes(
    Folder,
    {
        "parent": Folder,
        "name": str,
        "folders": List[Folder],
        "files": List[File],
    },
)


Entity.attributes(
    File,
    {
        "parent": Folder,
        "name": str,
        "content": str,
    },
)


def create_folder(name: str, _id: Optional[str] = None) -> Folder:
    result = Folder(_id=_id).rd_persist()
    result.name = name

    return result


class TestMultipleEntities(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        prepare_test()

    def test_model(self):
        @redino.connect
        def create_data() -> None:
            tests = create_folder("tests", _id="test")
            tests.folders = [
                create_folder("a"),
                create_folder("b"),
            ]

        @redino.connect
        def validate_data() -> None:
            tests = Folder(_id="test")

            self.assertEqual("tests", tests.name)
            self.assertEqual(2, len(tests.folders))
            self.assertEqual("a", tests.folders[0].name)
            self.assertEqual("b", tests.folders[1].name)

        @redino.connect
        def cleanup_data() -> None:
            Folder(_id="test").rd_delete()

        try:
            create_data()
            validate_data()
        finally:
            cleanup_data()


if __name__ == "__main__":
    unittest.main()
